@@ -1,4 +1,6 @@
-#   Copyright 2000-2004 Michael Hudson mwh@python.net
+#   Copyright 2000-2010 Michael Hudson-Doyle <micahel@gmail.com>
+#                       Antonio Cuni
+#                       Armin Rigo
 #
 #                        All Rights Reserved
 #
@@ -19,7 +21,6 @@
 
 import types, unicodedata
 from pyrepl import commands
-from curses import ascii
 from pyrepl import input
 
 def _make_unctrl_map():
@@ -30,6 +31,7 @@ def _make_unctrl_map():
     for i in range(32):
         c = unichr(i)
         uc_map[c] = u'^' + unichr(ord('A') + i - 1)
+    uc_map['\t'] = '    ' # display TABs as 4 characters
     uc_map['\177'] = u'^?'
     for i in range(256):
         c = unichr(i)
@@ -136,7 +138,7 @@ default_keymap = tuple(
      (r'\M-7', 'digit-arg'),
      (r'\M-8', 'digit-arg'),
      (r'\M-9', 'digit-arg'),
-     (r'\M-\n', 'insert-nl'),
+     #(r'\M-\n', 'insert-nl'),
      ('\\\\', 'self-insert')] + \
     [(c, 'self-insert')
      for c in map(chr, range(32, 127)) if c <> '\\'] + \
@@ -221,6 +223,8 @@ This is pyrepl.  Hear my roar.
 Helpful text may appear here at some point in the future when I'm
 feeling more loquacious than I am now."""
 
+    msg_at_bottom = True
+
     def __init__(self, console):
         self.buffer = []
         self.ps1 = "->> "
@@ -264,7 +268,7 @@ feeling more loquacious than I am now."""
         for ln, line in zip(range(len(lines)), lines):
             ll = len(line)
             if 0 <= p <= ll:
-                if self.msg:
+                if self.msg and not self.msg_at_bottom:
                     for mline in self.msg.split("\n"):
                         screen.append(mline)
                         screeninfo.append((0, []))
@@ -287,6 +291,10 @@ feeling more loquacious than I am now."""
                 screeninfo.append((0, l2[wrapcount*w - lp:]+[1]))
         self.screeninfo = screeninfo
         self.cxy = self.pos2xy(self.pos)
+        if self.msg and self.msg_at_bottom:
+            for mline in self.msg.split("\n"):
+                screen.append(mline)
+                screeninfo.append((0, []))
         return screen
 
     def bow(self, p=None):
@@ -460,6 +468,10 @@ feeling more loquacious than I am now."""
         self.dirty = 1
         self.console.beep()
 
+    def update_screen(self):
+        if self.dirty:
+            self.refresh()
+
     def refresh(self):
         """Recalculate and refresh the screen."""
         # this call sets up self.cxy, so call it first.
@@ -525,6 +537,10 @@ feeling more loquacious than I am now."""
 
             self.do_cmd(cmd)
             return 1
+
+    def push_char(self, char):
+        self.console.push_char(char)
+        self.handle1(0)
     
     def readline(self):
         """Read a line.  The implementation of this method also shows
