@@ -1,4 +1,6 @@
-#   Copyright 2000-2004 Michael Hudson mwh@python.net
+#   Copyright 2000-2010 Michael Hudson-Doyle <micahel@gmail.com>
+#                       Antonio Cuni
+#                       Armin Rigo
 #
 #                        All Rights Reserved
 #
@@ -17,13 +19,16 @@
 # CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 # CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
-import termios, curses, select, os, struct, errno
+import termios, select, os, struct, errno
 import signal, re, time, sys
 from fcntl import ioctl
+from pyrepl import curses
 from pyrepl.fancy_termios import tcgetattr, tcsetattr
 from pyrepl.console import Console, Event
 from pyrepl import unix_eventqueue
 
+_error = (termios.error, curses.error)
+
 # there are arguments for changing this to "refresh"
 SIGWINCH_EVENT = 'repaint'
 
@@ -81,10 +86,8 @@ class UnixConsole(Console):
             self.input_fd = f_in.fileno()
 
         if isinstance(f_out, int):
-            self.output = os.fdopen(f_out)
             self.output_fd = f_out
         else:
-            self.output = f_out
             self.output_fd = f_out.fileno()
         
         self.pollob = poll()
@@ -151,16 +154,18 @@ class UnixConsole(Console):
         self.__move = self.__move_short
 
         self.event_queue = unix_eventqueue.EventQueue(self.input_fd)
+        self.partial_char = ''
+        self.cursor_visible = 1
 
     def change_encoding(self, encoding):
         self.encoding = encoding
     
     def refresh(self, screen, (cx, cy)):
         # this function is still too long (over 90 lines)
-        self.__maybe_write_code(self._civis)
 
         if not self.__gone_tall:
             while len(self.screen) < min(len(screen), self.height):
+                self.__hide_cursor()
                 self.__move(0, len(self.screen) - 1)
                 self.__write("\n")
                 self.__posxy = 0, len(self.screen)
@@ -196,13 +201,14 @@ class UnixConsole(Console):
             offset = cy - height + 1
         elif offset > 0 and len(screen) < offset + height:
             offset = max(len(screen) - height, 0)
-            screen.append([])
+            screen.append("")
 
         oldscr = self.screen[old_offset:old_offset + height]
         newscr = screen[offset:offset + height]
 
         # use hardware scrolling if we have it.
         if old_offset > offset and self._ri:
+            self.__hide_cursor()
             self.__write_code(self._cup, 0, 0)
             self.__posxy = 0, old_offset
             for i in range(old_offset - offset):
@@ -210,6 +216,7 @@ class UnixConsole(Console):
                 oldscr.pop(-1)
                 oldscr.insert(0, "")
         elif old_offset < offset and self._ind:
+            self.__hide_cursor()
             self.__write_code(self._cup, self.height - 1, 0)
             self.__posxy = 0, old_offset + self.height - 1
             for i in range(offset - old_offset):
@@ -223,29 +230,34 @@ class UnixConsole(Console):
                                         oldscr,
                                         newscr):
             if oldline != newline:
-                self.write_changed_line(y, oldline, newline, px)
+                self.__write_changed_line(y, oldline, newline, px)
                 
         y = len(newscr)
         while y < len(oldscr):
+            self.__hide_cursor()
             self.__move(0, y)
             self.__posxy = 0, y
             self.__write_code(self._el)
             y += 1
 
-        self.__maybe_write_code(self._cnorm)
+        self.__show_cursor()
         
-        self.flushoutput()
         self.screen = screen
         self.move_cursor(cx, cy)
+        self.flushoutput()
 
-    def write_changed_line(self, y, oldline, newline, px):
+    def __write_changed_line(self, y, oldline, newline, px):
         # this is frustrating; there's no reason to test (say)
         # self.dch1 inside the loop -- but alternative ways of
         # structuring this function are equally painful (I'm trying to
         # avoid writing code generators these days...)
         x = 0
         minlen = min(len(oldline), len(newline))
-        while x < minlen and oldline[x] == newline[x]:
+        #
+        # reuse the oldline as much as possible, but stop as soon as we
+        # encounter an ESCAPE, because it might be the start of an escape
+        # sequene
+        while x < minlen and oldline[x] == newline[x] and newline[x] != '\x1b':
             x += 1
         if oldline[x:] == newline[x+1:] and self.ich1:
             if ( y == self.__posxy[1] and x > self.__posxy[0]
@@ -262,6 +274,7 @@ class UnixConsole(Console):
         elif (self.dch1 and self.ich1 and len(newline) == self.width
               and x < len(newline) - 2
               and newline[x+1:-1] == oldline[x:-2]):
+            self.__hide_cursor()
             self.__move(self.width - 2, y)
             self.__posxy = self.width - 2, y
             self.__write_code(self.dch1)
@@ -270,12 +283,12 @@ class UnixConsole(Console):
             self.__write(newline[x])
             self.__posxy = x + 1, y
         else:
+            self.__hide_cursor()
             self.__move(x, y)
             if len(oldline) > len(newline):
                 self.__write_code(self._el)
             self.__write(newline[x:])
             self.__posxy = len(newline), y
-        self.flushoutput()
 
     def __write(self, text):
         self.__buffer.append((text, 0))
@@ -376,35 +389,35 @@ class UnixConsole(Console):
         self.height, self.width = self.getheightwidth()
         self.event_queue.insert(Event('resize', None))
 
+    def push_char(self, char):
+        self.partial_char += char
+        try:
+            c = unicode(self.partial_char, self.encoding)
+        except UnicodeError, e:
+            if len(e.args) > 4 and \
+                   e.args[4] == 'unexpected end of data':
+                pass
+            else:
+                raise
+        else:
+            self.partial_char = ''
+            self.event_queue.push(c)
+        
     def get_event(self, block=1):
         while self.event_queue.empty():
             while 1: # All hail Unix!
                 try:
-                    c = ''
-                    while 1:
-                        c += os.read(self.input_fd, 1)
-                        try:
-                            c = unicode(c, self.encoding)
-                        except UnicodeError, e:
-                            if len(e.args) > 4 and \
-                               e.args[4] == 'unexpected end of data':
-                                continue
-                            else:
-                                raise
-                        else:
-                            break
-                        
-                except IOError, err:
+                    self.push_char(os.read(self.input_fd, 1))
+                except (IOError, OSError), err:
                     if err.errno == errno.EINTR:
                         if not self.event_queue.empty():
                             return self.event_queue.get()
                         else:
-                            continue # be explicit
+                            continue
                     else:
                         raise
                 else:
                     break
-            self.event_queue.push(c)
             if not block:
                 break
         return self.event_queue.get()
@@ -414,9 +427,19 @@ class UnixConsole(Console):
 
     def set_cursor_vis(self, vis):
         if vis:
-            self.__maybe_write_code(self._cnorm)
+            self.__show_cursor()
         else:
+            self.__hide_cursor()
+
+    def __hide_cursor(self):
+        if self.cursor_visible:
             self.__maybe_write_code(self._civis)
+            self.cursor_visible = 0
+
+    def __show_cursor(self):
+        if not self.cursor_visible:
+            self.__maybe_write_code(self._cnorm)
+            self.cursor_visible = 1
 
     def repaint_prep(self):
         if not self.__gone_tall:
